/*
 * linux/kernel/irq/proc.c
 *
 * Copyright (C) 1992, 1998-2004 Linus Torvalds, Ingo Molnar
 *
 * This file contains the /proc/irq/ handling code.
 */

#include <linux/irq.h>
#include <linux/proc_fs.h>
#include <linux/seq_file.h>
#include <linux/interrupt.h>
#include <linux/poll.h>

#include "internals.h"

static struct proc_dir_entry *root_irq_dir;

#ifdef CONFIG_SMP

static int irq_affinity_proc_show(struct seq_file *m, void *v)
{
	struct irq_desc *desc = irq_to_desc((long)m->private);
	cpumask_t *mask = &desc->affinity;

#ifdef CONFIG_GENERIC_PENDING_IRQ
	if (desc->status & IRQ_MOVE_PENDING)
		mask = &desc->pending_mask;
#endif
	seq_cpumask(m, mask);
	seq_putc(m, '\n');
	return 0;
}

#ifndef is_affinity_mask_valid
#define is_affinity_mask_valid(val) 1
#endif

int no_irq_affinity;
static ssize_t irq_affinity_proc_write(struct file *file,
		const char __user *buffer, size_t count, loff_t *pos)
{
	unsigned int irq = (int)(long)PDE(file->f_path.dentry->d_inode)->data;
	cpumask_t new_value;
	int err;

	if (!irq_to_desc(irq)->chip->set_affinity || no_irq_affinity ||
	    irq_balancing_disabled(irq))
		return -EIO;

	err = cpumask_parse_user(buffer, count, new_value);
	if (err)
		return err;

	if (!is_affinity_mask_valid(new_value))
		return -EINVAL;

	/*
	 * Do not allow disabling IRQs completely - it's a too easy
	 * way to make the system unusable accidentally :-) At least
	 * one online CPU still has to be targeted.
	 */
	if (!cpus_intersects(new_value, cpu_online_map))
		/* Special case for empty set - allow the architecture
		   code to set default SMP affinity. */
		return irq_select_affinity_usr(irq) ? -EINVAL : count;

	irq_set_affinity(irq, new_value);

	return count;
}

static int irq_affinity_proc_open(struct inode *inode, struct file *file)
{
	return single_open(file, irq_affinity_proc_show, PDE(inode)->data);
}

static const struct file_operations irq_affinity_proc_fops = {
	.open		= irq_affinity_proc_open,
	.read		= seq_read,
	.llseek		= seq_lseek,
	.release	= single_release,
	.write		= irq_affinity_proc_write,
};

static int default_affinity_show(struct seq_file *m, void *v)
{
	seq_cpumask(m, &irq_default_affinity);
	seq_putc(m, '\n');
	return 0;
}

static ssize_t default_affinity_write(struct file *file,
		const char __user *buffer, size_t count, loff_t *ppos)
{
	cpumask_t new_value;
	int err;

	err = cpumask_parse_user(buffer, count, new_value);
	if (err)
		return err;

	if (!is_affinity_mask_valid(new_value))
		return -EINVAL;

	/*
	 * Do not allow disabling IRQs completely - it's a too easy
	 * way to make the system unusable accidentally :-) At least
	 * one online CPU still has to be targeted.
	 */
	if (!cpus_intersects(new_value, cpu_online_map))
		return -EINVAL;

	irq_default_affinity = new_value;

	return count;
}

static int default_affinity_open(struct inode *inode, struct file *file)
{
	return single_open(file, default_affinity_show, NULL);
}

static const struct file_operations default_affinity_proc_fops = {
	.open		= default_affinity_open,
	.read		= seq_read,
	.llseek		= seq_lseek,
	.release	= single_release,
	.write		= default_affinity_write,
};
#endif

static int irq_spurious_read(char *page, char **start, off_t off,
				  int count, int *eof, void *data)
{
	struct irq_desc *desc = irq_to_desc((long) data);
	return sprintf(page, "count %u\n"
			     "unhandled %u\n"
			     "last_unhandled %u ms\n",
			desc->irq_count,
			desc->irqs_unhandled,
			jiffies_to_msecs(desc->last_unhandled));
}

#define MAX_NAMELEN 128

static int name_unique(unsigned int irq, struct irqaction *new_action)
{
	struct irq_desc *desc = irq_to_desc(irq);
	struct irqaction *action;
	unsigned long flags;
	int ret = 1;

	spin_lock_irqsave(&desc->lock, flags);
	for (action = desc->action ; action; action = action->next) {
		if ((action != new_action) && action->name &&
				!strcmp(new_action->name, action->name)) {
			ret = 0;
			break;
		}
	}
	spin_unlock_irqrestore(&desc->lock, flags);
	return ret;
}

void register_handler_proc(unsigned int irq, struct irqaction *action)
{
	char name [MAX_NAMELEN];
	struct irq_desc *desc = irq_to_desc(irq);

	if (!desc->dir || action->dir || !action->name ||
					!name_unique(irq, action))
		return;

	memset(name, 0, MAX_NAMELEN);
	snprintf(name, MAX_NAMELEN, "%s", action->name);

	/* create /proc/irq/1234/handler/ */
	action->dir = proc_mkdir(name, desc->dir);
}

#undef MAX_NAMELEN

#define MAX_NAMELEN 10

struct irq_proc {
        unsigned long irq;
        wait_queue_head_t q;
        atomic_t count;
        char devname[TASK_COMM_LEN];
};

static irqreturn_t irq_proc_irq_handler(int irq, void *vidp)
{
        struct irq_proc *idp = (struct irq_proc *)vidp;
        unsigned long stamp;

        BUG_ON(idp->irq != irq);

        disable_irq_nosync(irq);
        atomic_inc(&idp->count);

        wake_up(&idp->q);
        return IRQ_HANDLED;
}


/*
 * Signal to userspace an interrupt has occured.
 */
static ssize_t irq_proc_read(struct file *filp, char  __user *bufp, size_t len, loff_t *ppos)
{
        struct irq_proc *ip = (struct irq_proc *)filp->private_data;
        irq_desc_t *idp = irq_desc + ip->irq;
        int pending;

        DEFINE_WAIT(wait);

        if (len < sizeof(int))
                return -EINVAL;

        pending = atomic_read(&ip->count);
        if (pending == 0) {
                if (idp->status & IRQ_DISABLED)
                        enable_irq(ip->irq);
                if (filp->f_flags & O_NONBLOCK)
                        return -EWOULDBLOCK;
        }

        while (pending == 0) {
                prepare_to_wait(&ip->q, &wait, TASK_INTERRUPTIBLE);
                pending = atomic_read(&ip->count);
                if (pending == 0)
                        schedule();
                finish_wait(&ip->q, &wait);
                if (signal_pending(current))
                        return -ERESTARTSYS;
        }

        if (copy_to_user(bufp, &pending, sizeof pending))
                return -EFAULT;

        *ppos += sizeof pending;

        atomic_sub(pending, &ip->count);
        return sizeof pending;
}


static int irq_proc_open(struct inode *inop, struct file *filp)
{
        struct irq_proc *ip;
        struct proc_dir_entry *ent = PDE(inop);
        int error;

        ip = kmalloc(sizeof *ip, GFP_KERNEL);
        if (ip == NULL)
                return -ENOMEM;

        memset(ip, 0, sizeof(*ip));
        strcpy(ip->devname, current->comm);
        init_waitqueue_head(&ip->q);
        atomic_set(&ip->count, 0);
        ip->irq = (unsigned long)ent->data;

        error = request_irq(ip->irq,
                            irq_proc_irq_handler,
                            0,
                            ip->devname,
                            ip);
        if (error < 0) {
                kfree(ip);
                return error;
        }
        filp->private_data = (void *)ip;

        return 0;
}

static int irq_proc_release(struct inode *inop, struct file *filp)
{
        struct irq_proc *ip = (struct irq_proc *)filp->private_data;

        free_irq(ip->irq, ip);
        filp->private_data = NULL;
        kfree(ip);
        return 0;
}

static unsigned int irq_proc_poll(struct file *filp, struct poll_table_struct *wait)
{
        struct irq_proc *ip = (struct irq_proc *)filp->private_data;
        irq_desc_t *idp = irq_desc + ip->irq;

        if (atomic_read(&ip->count) > 0)
                return POLLIN | POLLRDNORM; /* readable */

        /* if interrupts disabled and we don't have one to process... */
        if (idp->status & IRQ_DISABLED)
                enable_irq(ip->irq);

        poll_wait(filp, &ip->q, wait);

        if (atomic_read(&ip->count) > 0)
                return POLLIN | POLLRDNORM; /* readable */

        return 0;
}

static struct file_operations irq_proc_file_operations = {
        .read = irq_proc_read,
        .open = irq_proc_open,
        .release = irq_proc_release,
        .poll = irq_proc_poll,
};

void register_irq_proc(unsigned int irq, struct irq_desc *desc)
{
	struct proc_dir_entry *entry;
	struct proc_dir_entry *entry_sp;
	char name [MAX_NAMELEN];

//	if (!root_irq_dir || (desc->chip == &no_irq_chip) || desc->dir)
	if (!root_irq_dir)
		return;

	memset(name, 0, MAX_NAMELEN);
	sprintf(name, "%d", irq);

	/* create /proc/irq/1234 */
//	desc->dir = proc_mkdir(name, root_irq_dir);

        if (!desc->dir) {
                /* create /proc/irq/1234 */
                desc->dir = proc_mkdir(name, root_irq_dir);
 
                /*
                 * Create handles for user-mode interrupt handlers
                 * if the kernel hasn't already grabbed the IRQ
                 */
                entry = create_proc_entry("irq", 0600, desc->dir);
                if (entry) {
                        entry->data = (void *)(unsigned long)irq;
			entry->read_proc = NULL;
                        entry->write_proc = NULL;
                        entry->proc_fops = &irq_proc_file_operations;
                }
	
		entry_sp = create_proc_entry("spurious", 0444, desc->dir);
	        if (entry_sp) {
        	        entry_sp->data = (void *)(long)irq;
	                entry_sp->read_proc = irq_spurious_read;
        	}
        }

#ifdef CONFIG_SMP
	/* create /proc/irq/<irq>/smp_affinity */
	proc_create_data("smp_affinity", 0600, desc->dir,
			 &irq_affinity_proc_fops, (void *)(long)irq);
#endif
}

#undef MAX_NAMELEN

void unregister_handler_proc(unsigned int irq, struct irqaction *action)
{
	if (action->dir) {
		struct irq_desc *desc = irq_to_desc(irq);

		remove_proc_entry(action->dir->name, desc->dir);
	}
}

static void register_default_affinity_proc(void)
{
#ifdef CONFIG_SMP
	proc_create("irq/default_smp_affinity", 0600, NULL,
		    &default_affinity_proc_fops);
#endif
}

void init_irq_proc(void)
{
	unsigned int irq;
	struct irq_desc *desc;

	/* create /proc/irq */
	root_irq_dir = proc_mkdir("irq", NULL);
	if (!root_irq_dir)
		return;

	register_default_affinity_proc();

	/*
	 * Create entries for all existing IRQs.
	 */
	for_each_irq_desc(irq, desc)
		register_irq_proc(irq, desc);
}

